# Lab 02: Sanitization

TL;DR The goal of this lab is to get familiar with sanitization. You will set up
the Juliet test suite on your machine and run a set of sanitizers on them to
learn what kind of CWE's each sanitizer protects against and what the
limitations of each sanitizer is.

Please collect and write up the results for each task in a short report. For
each task, you should invest about 2 hours of thinking time (and execute the
tests alongside). Send your (short) report to <mailto:mathias.payer@epfl.ch> by
December 4 (in 3 weeks).


## Background: Sanitization

Sanitizers enforce a security (or safety policy) and make some form of undefined
behavior explicit. Instead of silently corrupting program state, the sanitizer
detects the violation when it happens and stops program execution with a
detailed error report. Sanitizers result in high overhead due to the additional
metadata they track and therefore are a development tool, often run with test
cases or with fuzzing.

We will evaluate a test suite with three different sanitizers that each enforce
different policies: ASan, MSan, and UBSan:

* *AddressSanitizer* (ASan) detects spatial and some temporal memory safety
  violations by instrumenting object allocations/deallocations and pointer
  dereferences. ASan keeps metadata about each allocated memory object,
  including where the object was allocated and if the object is still valid.
  Each pointer read or write is instrumented with fast validity check and
  violations trigger an exception and a report. You can compile your code with
  ASan through, e.g., `clang -fsanitize=address prog.c`. For more details see
  the full 
  [ASan description](https://github.com/google/sanitizers/wiki/AddressSanitizer).

* *MemorySanitizer* (MSan) detects uninitialized reads by instrumenting object
  allocations/deallocations and pointer dereferences. MSan keeps metadata about
  each allocated object and remembers *which* bytes have already been written.
  Each pointer read is instrumented to check if the current data is valid. Each
  pointer write is instrumented to tag the underlying area as written. You can
  compile your code with MSan through, e.g., `clang -fsanitize=memory prog.c`.
  For more details see the full 
  [MSan description](https://clang.llvm.org/docs/MemorySanitizer.html).

* *UndefinedBehaviorSanitizer* (UBSan) detects generic undefined behavior such
  as unaligned pointer dereferences, null pointer accesses, signed integer
  overflow, or floating point conversion (depending on the compiler flags). The
  sanitizer adds instrumentation to enforce checks for the different properties
  and keeps the required metadata. You can compileryour code with UBSan through,
  e.g., `clang -fsanitize=undefined prog.c`. For more details see the full
  full 
  [UBSan description](https://clang.llvm.org/docs/UndefinedBehaviorSanitizer.html).


## Task 0: set-up your sanitization environment

The first step is to set up the necessary build environment. You'll need at
least a recent version of clang installed (4.0 should suffice). LLVM comes with
a set of sanitizers by default which allow you to compile programs with
additional sanitization policies. Sanitizers are exclusive, so you can only use
one sanitizer at a time. We will leverage three different sanitizers: ASan,
MSan, and UBSan on a test suite with a lot of flaws.

The test suite we will use for this experiment is the 
[Juliet test suite](https://www.nist.gov/publications/juliet-11-cc-and-java-test-suite). 
You can download the test programs from the 
[NIST website](https://samate.nist.gov/SRD/testsuite.php). We will leverage the
updated version 1.3 of the Juliet Test Suite for C/C++ programs with 64,099 test
programs. Go ahead and download the test programs.

For the lazy: `./setup.sh` then rock and roll. Note that
`create_single_asan_Makefile.py` contains an example to compile the set of test
cases for a single CWE with ASan. 


## Task {1, 2, 3}: Evaluating {ASan, MSan, UBSan}

In these three tasks, we test the different sanitizer policies. For each 

* What CWE classes can the sanitizer theoretically detect? Read the description
  of the sanitizers and discuss which CWE classes are covered.

* Compile the tests with the sanitizer.

* Run the test cases and discuss what how many false positives (the sanitizer
  reports a violation even though the program is legal) and false negatives (the
  sanitizer misses a violation) you have. What could be the reason for the false
  positives or false negatives? Discuss an example of a false positive/negative.

* Play with the configuration of the sanitizer. How can you increase the true
  positive rate? How can you reduce the false positive rate? What flags work
  best?

* Create a summary statistic across the different sanitizers and configurations
  regarding false positives and negatives.


## Task 4 (voluntary, no credit): Missing sanitization?

For the curious and adventurous, look at the existing set of sanitizers in LLVM
and published at research conferences and at the CWEs. Discuss which CWEs are
not checked. How would you develop a sanitizer that covers these policies? Are
they enforceable?

